/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QString>
#include <vector>
#include <assert.h>

#include "correctiondata.h"

#include "../../core/interpolation.h"

static const double						MINIMUM_TEMPERATURE_RANGE(1.0);


TCorrectionData::TCorrectionData(void)
{
	d_data_pnts[0].Set(15,0);
	d_data_pnts[1].Set(20,0);
	d_data_pnts[2].Set(25,0);
	
	this->Update_Coefficients();
}

TCorrectionData::TCorrectionData(
	const TVector2 						&p1,
	const TVector2 						&p2,
	const TVector2 						&p3)
{
	d_data_pnts[0] = p1;
	d_data_pnts[1] = p2;
	d_data_pnts[2] = p3;
	
	this->Update_Coefficients();
}

TCorrectionData::TCorrectionData(
	const TCorrectionData 				&rhs)
{
	if(&rhs != this)
	{
		memcpy(d_data_pnts,rhs.d_data_pnts,sizeof(TVector2) * 3);
		memcpy(d_data_coeff,rhs.d_data_coeff,sizeof(double) * 3);
	}
}

TCorrectionData& TCorrectionData::operator=(
	const TCorrectionData 				&rhs)
{
	if(&rhs != this)
	{
		memcpy(d_data_pnts,rhs.d_data_pnts,sizeof(TVector2) * 3);
		memcpy(d_data_coeff,rhs.d_data_coeff,sizeof(double) * 3);
	}
	
	return *this;
}

TCorrectionData::~TCorrectionData(void)
{
}

TVector2 TCorrectionData::Value(
	const double						&pos) const
{
	TVector2							pnt;
	
	pnt.x = pos;
	
	pnt.y = d_data_coeff[2];
	pnt.y += (d_data_coeff[1] * pos);
	pnt.y += (d_data_coeff[0] * (pos * pos));
	
	return pnt;
}

bool TCorrectionData::Valid_Range(void) const
{
	if((d_data_pnts[2].x - d_data_pnts[0].x) < MINIMUM_TEMPERATURE_RANGE)
	{
		return false;
	}
	
	return true;
}

void TCorrectionData::Set_Points(
	const TVector2						&p1,
	const TVector2						&p2,
	const TVector2						&p3)
{
	d_data_pnts[0] = p1;
	d_data_pnts[1] = p2;
	d_data_pnts[2] = p3;
	
	this->Update_Coefficients();
}

void TCorrectionData::Set_Point1(
	const TVector2						&pnt)
{
	d_data_pnts[0] = pnt;
	
	this->Update_Coefficients();
}

void TCorrectionData::Set_Point2(
	const TVector2						&pnt)
{
	d_data_pnts[1] = pnt;
	
	this->Update_Coefficients();
}

void TCorrectionData::Set_Point3(
	const TVector2						&pnt)
{
	d_data_pnts[2] = pnt;
	
	this->Update_Coefficients();
}

void TCorrectionData::Update_Coefficients(void)
{
	std::vector<TVector2>				data_pnts;
	TVector2							pnt;
	
	// sort the three points in accending order
	
	if(d_data_pnts[0].x > d_data_pnts[1].x)
	{
		pnt = d_data_pnts[0];
		d_data_pnts[0] = d_data_pnts[1];
		d_data_pnts[1] = pnt;
	}
	
	if(d_data_pnts[1].x > d_data_pnts[2].x)
	{
		pnt = d_data_pnts[1];
		d_data_pnts[1] = d_data_pnts[2];
		d_data_pnts[2] = pnt;
	}
	
	if(d_data_pnts[0].x > d_data_pnts[1].x)
	{
		pnt = d_data_pnts[0];
		d_data_pnts[0] = d_data_pnts[1];
		d_data_pnts[1] = pnt;
	}
	
	// if data range is below epsilon assume all coefficients are zero

	if((d_data_pnts[2].x - d_data_pnts[0].x) < MINIMUM_TEMPERATURE_RANGE)
	{
		d_data_coeff[0] = 0.0;
		d_data_coeff[1] = 0.0;
		d_data_coeff[2] = 0.0;
		
		return;
	}
	
	data_pnts.resize(3);
	data_pnts[0] = d_data_pnts[0];
	data_pnts[1] = d_data_pnts[1];
	data_pnts[2] = d_data_pnts[2];

	if(!LeastSquares::Fit_Curve(data_pnts,&d_data_coeff[0],&d_data_coeff[1],&d_data_coeff[2]))
	{
		d_data_coeff[0] = 0.0;
		d_data_coeff[1] = 0.0;
		d_data_coeff[2] = 0.0;
	}
}
