/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QAction>
#include <QApplication>
#include <QCheckBox>
#include <QComboBox>
#include <QFont>
#include <QFrame>
#include <QFile>
#include <QFileDialog>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QMenu>
#include <QMenuBar>
#include <QPlainTextEdit>
#include <QPushButton>
#include <QSizePolicy>
#include <QSpacerItem>
#include <QStringList>
#include <QToolButton>
#include <QVBoxLayout>
#include <QWidget>
#include <QTextCursor>
#include <QSettings>
#include <QPixmap>
#include <QCloseEvent>
#include <QDir>
#include <QStringList>
#include <cmath>
#include <assert.h>

#include "../../core/messagebox.h"

#include "correctiondialog.h"
#include "inputdialog.h"

#include "temperaturerecord.h"

TTemperatureRecord::TTemperatureRecord(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QMainWindow(const_cast<QWidget*>(parent),flags)
{
	QFont								reading_font;
	QFont								log_font;
	QFont								copyright_font;
	QFrame								*separator_hline1;
	QFrame								*separator_hline2;
	QFrame								*separator_hline3;
	QGridLayout							*entry_grid;
	QGridLayout							*record_grid;
	QHBoxLayout							*copyright_hlayout;
	QLabel								*copyright_label;
	QLabel								*copyright_logo;
	QLabel								*corrected_reading_label;
	QLabel								*log_label;
	QLabel								*meter_reading_label;
	QLabel								*record_file_label;
	QLabel								*record_path_label;
	QLabel								*thermometer_label;
	QLabel								*description_label;
	QMenu								*edit_menu;
	QMenu								*file_menu;
	QMenuBar							*menubar;
	QSpacerItem							*verticalSpacer;
	QVBoxLayout							*central_widget_vlayout;
	QWidget								*central_widget;
	QSizePolicy							sizepolicy_expanding_preferred(QSizePolicy::Expanding,QSizePolicy::Preferred);
	QDir								data_dir;
	QString								file_name;
	QString								text;
	QStringList							list;
	QStringList::const_iterator			iter;
	int									index;
	
	sizepolicy_expanding_preferred.setHorizontalStretch(0);
	sizepolicy_expanding_preferred.setVerticalStretch(0);

	reading_font.setPointSize(16);
	reading_font.setWeight(75);
	reading_font.setBold(true);

	log_font.setFamily(QStringLiteral("Courier New"));
	log_font.setFixedPitch(true);
	
	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);
	
	d_application_path = qApp->applicationDirPath();
	
#ifdef Q_OS_MAC
	index = d_application_path.lastIndexOf(QStringLiteral(".app/"));
	d_application_path.truncate(index);
	
	index = d_application_path.lastIndexOf('/');
	d_application_path.truncate(index);
#endif
	
#ifdef Q_OS_WIN
	d_data_path = d_application_path;
#else
	d_data_path = QDir::homePath();
#endif
	
	data_dir.setPath(d_data_path);
	
	if(!data_dir.exists(QStringLiteral(".temperaturerecord")))
	{
		data_dir.mkdir(QStringLiteral(".temperaturerecord"));
	}
	
	data_dir.cd(QStringLiteral(".temperaturerecord"));
	d_data_path = data_dir.absolutePath();
	
	d_settings = new QSettings(d_data_path + QStringLiteral("/temperaturerecord_settings.ini"),QSettings::IniFormat,this);

	this->resize(754,629);

	d_action_file_quit = new QAction(this);
	d_action_edit_thermometer = new QAction(this);
	d_action_remove_thermometer = new QAction(this);

	central_widget = new QWidget(this);
	this->setCentralWidget(central_widget);

	central_widget_vlayout = new QVBoxLayout(central_widget);

	record_grid = new QGridLayout();

	record_path_label = new QLabel(central_widget);
	record_grid->addWidget(record_path_label,0,0,1,1);

	d_record_path_edit = new QLineEdit(central_widget);
	record_grid->addWidget(d_record_path_edit,0,1,1,1);

	d_path_browse_button = new QToolButton(central_widget);
	record_grid->addWidget(d_path_browse_button,0,2,1,2);

	record_file_label = new QLabel(central_widget);
	record_grid->addWidget(record_file_label,1,0,1,1);

	d_record_file_edit = new QLineEdit(central_widget);
	record_grid->addWidget(d_record_file_edit,1,1,1,1);

	thermometer_label = new QLabel(central_widget);
	record_grid->addWidget(thermometer_label,2,0,1,1);
	
	d_thermometer_combo = new QComboBox(central_widget);
	record_grid->addWidget(d_thermometer_combo,2,1,1,1);
	
	d_thermometer_edit_button = new QToolButton(central_widget);
	record_grid->addWidget(d_thermometer_edit_button,2,2,1,1);
	
	d_thermometer_delete_button = new QToolButton(central_widget);
	record_grid->addWidget(d_thermometer_delete_button,2,3,1,1);

	central_widget_vlayout->addLayout(record_grid);

	separator_hline1 = new QFrame(central_widget);
	separator_hline1->setFrameShape(QFrame::HLine);
	separator_hline1->setFrameShadow(QFrame::Sunken);
	central_widget_vlayout->addWidget(separator_hline1);

	entry_grid = new QGridLayout();
	
	description_label = new QLabel(central_widget);
	description_label->setAlignment(Qt::AlignRight | Qt::AlignVCenter);
	entry_grid->addWidget(description_label,0,0,1,1);

	d_description_edit = new QLineEdit(central_widget);
	entry_grid->addWidget(d_description_edit,0,1,1,2);

	meter_reading_label = new QLabel(central_widget);
	meter_reading_label->setAlignment(Qt::AlignCenter);
	entry_grid->addWidget(meter_reading_label,1,1,1,1);

	corrected_reading_label = new QLabel(central_widget);
	corrected_reading_label->setAlignment(Qt::AlignCenter);
	entry_grid->addWidget(corrected_reading_label,1,2,1,1);

	d_temperature_x_check = new QCheckBox(central_widget);
	entry_grid->addWidget(d_temperature_x_check,2,0,1,1);

	d_meter_x_edit = new QLineEdit(central_widget);
	d_meter_x_edit->setFont(reading_font);
	d_meter_x_edit->setAlignment(Qt::AlignCenter);
	entry_grid->addWidget(d_meter_x_edit,2,1,1,1);

	d_corrected_x_edit = new QLineEdit(central_widget);
	d_corrected_x_edit->setFont(reading_font);
	d_corrected_x_edit->setAlignment(Qt::AlignCenter);
	d_corrected_x_edit->setReadOnly(true);
	entry_grid->addWidget(d_corrected_x_edit,2,2,1,1);
	
	d_write_button = new QPushButton(central_widget);
	d_write_button->setAutoDefault(false);
	d_write_button->setSizePolicy(sizepolicy_expanding_preferred);
	entry_grid->addWidget(d_write_button,2,3,2,1);

	d_evaluate_corrections_button = new QPushButton(central_widget);
	d_evaluate_corrections_button->setAutoDefault(false);
	d_evaluate_corrections_button->setSizePolicy(sizepolicy_expanding_preferred);
	entry_grid->addWidget(d_evaluate_corrections_button,4,3,1,1);
	
	d_evaluate_uncertainty_button = new QPushButton(central_widget);
	d_evaluate_uncertainty_button->setAutoDefault(false);
	d_evaluate_uncertainty_button->setSizePolicy(sizepolicy_expanding_preferred);
	entry_grid->addWidget(d_evaluate_uncertainty_button,5,3,1,1);

	d_temperature_y_check = new QCheckBox(central_widget);
	entry_grid->addWidget(d_temperature_y_check,3,0,1,1);

	d_meter_y_edit = new QLineEdit(central_widget);
	d_meter_y_edit->setFont(reading_font);
	d_meter_y_edit->setAlignment(Qt::AlignCenter);
	entry_grid->addWidget(d_meter_y_edit,3,1,1,1);

	d_corrected_y_edit = new QLineEdit(central_widget);
	d_corrected_y_edit->setFont(reading_font);
	d_corrected_y_edit->setAlignment(Qt::AlignCenter);
	d_corrected_y_edit->setReadOnly(true);
	entry_grid->addWidget(d_corrected_y_edit,3,2,1,1);

	d_temperature_z_check = new QCheckBox(central_widget);
	entry_grid->addWidget(d_temperature_z_check,4,0,1,1);

	d_meter_z_edit = new QLineEdit(central_widget);
	d_meter_z_edit->setFont(reading_font);
	d_meter_z_edit->setAlignment(Qt::AlignCenter);
	entry_grid->addWidget(d_meter_z_edit,4,1,1,1);

	d_corrected_z_edit = new QLineEdit(central_widget);
	d_corrected_z_edit->setFont(reading_font);
	d_corrected_z_edit->setAlignment(Qt::AlignCenter);
	d_corrected_z_edit->setReadOnly(true);
	entry_grid->addWidget(d_corrected_z_edit,4,2,1,1);

	d_temperature_part_check = new QCheckBox(central_widget);
	entry_grid->addWidget(d_temperature_part_check,5,0,1,1);

	d_meter_part_edit = new QLineEdit(central_widget);
	d_meter_part_edit->setFont(reading_font);
	d_meter_part_edit->setAlignment(Qt::AlignCenter);
	entry_grid->addWidget(d_meter_part_edit,5,1,1,1);

	d_corrected_part_edit = new QLineEdit(central_widget);
	d_corrected_part_edit->setFont(reading_font);
	d_corrected_part_edit->setAlignment(Qt::AlignCenter);
	d_corrected_part_edit->setReadOnly(true);
	entry_grid->addWidget(d_corrected_part_edit,5,2,1,1);
	central_widget_vlayout->addLayout(entry_grid);

	separator_hline2 = new QFrame(central_widget);
	separator_hline2->setFrameShape(QFrame::HLine);
	separator_hline2->setFrameShadow(QFrame::Sunken);
	central_widget_vlayout->addWidget(separator_hline2);

	log_label = new QLabel(central_widget);
	central_widget_vlayout->addWidget(log_label);

	d_log_edit = new QPlainTextEdit(central_widget);
	d_log_edit->setFont(log_font);
	d_log_edit->setLineWrapMode(QPlainTextEdit::NoWrap);
	d_log_edit->setReadOnly(true);
	central_widget_vlayout->addWidget(d_log_edit);

	separator_hline3 = new QFrame(central_widget);
	separator_hline3->setFrameShape(QFrame::HLine);
	separator_hline3->setFrameShadow(QFrame::Sunken);
	central_widget_vlayout->addWidget(separator_hline3);

	verticalSpacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
	central_widget_vlayout->addItem(verticalSpacer);

	copyright_hlayout = new QHBoxLayout();

	copyright_label = new QLabel(central_widget);
	copyright_label->setFont(copyright_font);
	copyright_label->setAlignment(Qt::AlignBottom|Qt::AlignLeading|Qt::AlignLeft);
	copyright_hlayout->addWidget(copyright_label);

	copyright_logo = new QLabel(central_widget);
	copyright_logo->setFixedSize(QSize(100,40));
	copyright_hlayout->addWidget(copyright_logo);
	central_widget_vlayout->addLayout(copyright_hlayout);

	menubar = new QMenuBar(this);
	this->setMenuBar(menubar);

	file_menu = new QMenu(menubar);
	menubar->addMenu(file_menu);
	file_menu->addAction(d_action_file_quit);

	edit_menu = new QMenu(menubar);
	menubar->addMenu(edit_menu);
	edit_menu->addAction(d_action_edit_thermometer);
	edit_menu->addAction(d_action_remove_thermometer);

	menubar->addAction(file_menu->menuAction());
	menubar->addAction(edit_menu->menuAction());
	
	// defaults		
	d_correction_dialog = 0;
	d_input_dialog = 0;
	d_msg_box = new TMessageBox(this);
	d_active_thermometer_usable = false;
	
	d_current_path = d_settings->value("Current_Path", QDir::homePath()).toString();
	file_name = d_settings->value("Record_File", QStringLiteral("temperature_record.csv")).toString();

	d_record_path_edit->setText(QDir::toNativeSeparators(d_current_path));
	d_record_file_edit->setText(file_name);
	
	file_name = d_data_path + QStringLiteral("/calibration_data.xml");
	
	if(d_thermometer_data.Load_Data(file_name))
	{
		this->Add_Log_Text(QStringLiteral("INF  Calibration file loaded: '%1'").arg(QDir::toNativeSeparators(file_name)));
	}
	else
	{
		this->Add_Log_Text(QStringLiteral("ERR  Cannot load calibration file: '%1'").arg(QDir::toNativeSeparators(file_name)));
	}
	
	list = d_thermometer_data.Description_List();
	d_thermometer_combo->addItems(list);
	
	for(iter = list.begin();iter != list.end();++iter)
	{
		text = d_thermometer_data.Get_Calibration_Due_Date(*iter).toString(QStringLiteral("MMM dd yyyy"));
		this->Add_Log_Text(QString("INF  Loaded correction data for '%1'.  Calibration Due: %2").arg(*iter).arg(text));
	}
	
	this->Add_Log_Text(QString("INF  Calibration data checksum error count: %1").arg(d_thermometer_data.Checksum_Error_Count()));

	d_description_edit->setText(d_settings->value("Description",QStringLiteral("Position 1")).toString());
	
	d_temperature_x_check->setChecked(d_settings->value("Axis_X_Enabled",true).toBool());
	d_temperature_y_check->setChecked(d_settings->value("Axis_Y_Enabled",true).toBool());
	d_temperature_z_check->setChecked(d_settings->value("Axis_Z_Enabled",true).toBool());
	d_temperature_part_check->setChecked(d_settings->value("Axis_Part_Enabled",true).toBool());
	
	d_meter_x_edit->setText(d_settings->value("Meter_X_Value",QStringLiteral("20.00")).toString());
	d_meter_y_edit->setText(d_settings->value("Meter_Y_Value",QStringLiteral("20.00")).toString());
	d_meter_z_edit->setText(d_settings->value("Meter_Z_Value",QStringLiteral("20.00")).toString());
	d_meter_part_edit->setText(d_settings->value("Meter_Part_Value",QStringLiteral("20.00")).toString());

	QWidget::setTabOrder(d_record_path_edit,d_path_browse_button);
	QWidget::setTabOrder(d_path_browse_button,d_record_file_edit);
	QWidget::setTabOrder(d_record_file_edit,d_thermometer_combo);
	QWidget::setTabOrder(d_thermometer_combo,d_thermometer_edit_button);
	QWidget::setTabOrder(d_thermometer_edit_button,d_thermometer_delete_button);
	QWidget::setTabOrder(d_thermometer_delete_button,d_description_edit);
	QWidget::setTabOrder(d_description_edit,d_temperature_x_check);
	QWidget::setTabOrder(d_temperature_x_check,d_meter_x_edit);
	QWidget::setTabOrder(d_meter_x_edit,d_temperature_y_check);
	QWidget::setTabOrder(d_temperature_y_check,d_meter_y_edit);
	QWidget::setTabOrder(d_meter_y_edit,d_temperature_z_check);
	QWidget::setTabOrder(d_temperature_z_check,d_meter_z_edit);
	QWidget::setTabOrder(d_meter_z_edit,d_temperature_part_check);
	QWidget::setTabOrder(d_temperature_part_check,d_meter_part_edit);
	QWidget::setTabOrder(d_meter_part_edit,d_write_button);
	QWidget::setTabOrder(d_write_button,d_evaluate_corrections_button);
	QWidget::setTabOrder(d_evaluate_corrections_button,d_evaluate_uncertainty_button);
	QWidget::setTabOrder(d_evaluate_uncertainty_button,d_log_edit);

	this->setWindowTitle(QStringLiteral("Temperature Record - Version 1.1"));
	
	copyright_logo->setPixmap(QPixmap(QStringLiteral(":/images/logo.png")));
	
	d_thermometer_edit_button->setIconSize(QSize(16,16));
	d_thermometer_delete_button->setIconSize(QSize(16,16));
	d_path_browse_button->setIconSize(QSize(16,16));

	d_thermometer_edit_button->setIcon(QIcon(QStringLiteral(":/gui/edit_icon16.png")));
	d_thermometer_delete_button->setIcon(QIcon(QStringLiteral(":/gui/remove_icon16.png")));
	d_path_browse_button->setIcon(QIcon(QStringLiteral(":/gui/browse_icon16.png")));

	d_action_file_quit->setText(QStringLiteral("Quit"));
	d_action_edit_thermometer->setText(QStringLiteral("Correction Editor"));
	d_action_remove_thermometer->setText(QStringLiteral("Remove Thermometer"));
	record_path_label->setText(QStringLiteral("Record Path:"));
	record_file_label->setText(QStringLiteral("Record File:"));
	description_label->setText(QStringLiteral("Description:"));
	thermometer_label->setText(QStringLiteral("Thermometer:"));
	meter_reading_label->setText(QStringLiteral("Meter Reading ˚C"));
	corrected_reading_label->setText(QStringLiteral("Corrected Reading ˚C"));
	d_temperature_x_check->setText(QStringLiteral("X Axis Temperature:"));
	d_temperature_y_check->setText(QStringLiteral("Y Axis Temperature:"));
	d_temperature_z_check->setText(QStringLiteral("Z Axis Temperature:"));
	d_temperature_part_check->setText(QStringLiteral("Part Temperature:"));
	d_write_button->setText(QStringLiteral("Write"));
	d_evaluate_corrections_button->setText(QStringLiteral("Evaluate Corrections"));
	d_evaluate_uncertainty_button->setText(QStringLiteral("Evaluate Uncertainty"));
	log_label->setText(QStringLiteral("Log:"));
	copyright_label->setText(QStringLiteral("Copyright (C) 2023 Select Calibration Incorporated.  www.selectcalibration.ca"));
	file_menu->setTitle(QStringLiteral("File"));
	edit_menu->setTitle(QStringLiteral("Edit"));
	
	index = d_settings->value("Active_Thermometer",0).toInt();
	if(!(index < 0) && index < d_thermometer_combo->count())
	{
		d_thermometer_combo->setCurrentIndex(index);
		this->Thermometer_Selection_Changed(index);
	}
	else if(d_thermometer_combo->count())
	{
		d_thermometer_combo->setCurrentIndex(0);
		this->Thermometer_Selection_Changed(0);
	}
	else
	{
		this->Update_All_Temperature_Corrections();
	}
	
	connect(d_action_file_quit,&QAction::triggered,this,&TTemperatureRecord::close);
	connect(d_action_edit_thermometer,&QAction::triggered,this,&TTemperatureRecord::Edit_Thermometer);
	connect(d_action_remove_thermometer,&QAction::triggered,this,&TTemperatureRecord::Remove_Thermometer);
	connect(d_path_browse_button,&QToolButton::clicked,this,&TTemperatureRecord::Browse_Path);
	connect(d_thermometer_edit_button,&QToolButton::clicked,this,&TTemperatureRecord::Edit_Thermometer);
	connect(d_thermometer_delete_button,&QToolButton::clicked,this,&TTemperatureRecord::Remove_Thermometer);
	connect(d_thermometer_combo,static_cast<void (QComboBox::*)(int)>(&QComboBox::activated),this,&TTemperatureRecord::Thermometer_Selection_Changed);
	connect(d_meter_x_edit,&QLineEdit::editingFinished,this,&TTemperatureRecord::Temperature_Input_Changed);
	connect(d_meter_y_edit,&QLineEdit::editingFinished,this,&TTemperatureRecord::Temperature_Input_Changed);
	connect(d_meter_z_edit,&QLineEdit::editingFinished,this,&TTemperatureRecord::Temperature_Input_Changed);
	connect(d_meter_part_edit,&QLineEdit::editingFinished,this,&TTemperatureRecord::Temperature_Input_Changed);

	connect(d_write_button,&QToolButton::clicked,this,&TTemperatureRecord::Write_Temperature_Entry);
	connect(d_evaluate_corrections_button,&QToolButton::clicked,this,&TTemperatureRecord::Evaluate_Corrections);
	connect(d_evaluate_uncertainty_button,&QToolButton::clicked,this,&TTemperatureRecord::Evaluate_Uncertainty);
}

TTemperatureRecord::~TTemperatureRecord(void)
{
}

void TTemperatureRecord::Browse_Path(void)
{
	QString								file_path;
	
	file_path = QFileDialog::getExistingDirectory(this,
											 QStringLiteral("Select directory to write output file."),
											 QDir::fromNativeSeparators(d_record_path_edit->text()));
	
	if(file_path.length())
	{
		d_record_path_edit->setText(QDir::toNativeSeparators(file_path));
		d_current_path = file_path;
	}
}
	
void TTemperatureRecord::Edit_Thermometer(void)
{
	QStringList							description_list;
	QString								text;
	TCorrectionDialog::TDialogData		dialog_data;
	int									index;
	
	if(!d_correction_dialog)
	{
		d_correction_dialog = new TCorrectionDialog(this);
	}
	
	description_list = d_thermometer_data.Description_List();
	
	if(d_thermometer_combo->count())
	{
		dialog_data.description = d_thermometer_combo->currentText();
		description_list.removeOne(dialog_data.description);
	}
	
	d_thermometer_data.Get_Temperature_Data(dialog_data.description,&dialog_data.temperature_data);
	dialog_data.calibration_due_date = d_thermometer_data.Get_Calibration_Due_Date(dialog_data.description);
	
	d_correction_dialog->Reset(dialog_data,description_list);
	
	if(QDialog::Accepted == d_correction_dialog->exec())
	{
		d_correction_dialog->Dialog_Data(&dialog_data);
		
		d_thermometer_data.Set_Temperature_Data(dialog_data.description,dialog_data.temperature_data);
		d_thermometer_data.Set_Calibration_Due_Date(dialog_data.description,dialog_data.calibration_due_date);
		
		description_list = d_thermometer_data.Description_List();
		d_thermometer_combo->clear();
		d_thermometer_combo->addItems(description_list);
		
		index = d_thermometer_combo->findText(dialog_data.description);
		
		if(!(index < 0))
		{
			d_thermometer_combo->setCurrentIndex(index);
		}
		
		d_active_thermometer_usable = d_thermometer_data.Is_Calibration_Current(dialog_data.description);
		d_active_correction_data = d_thermometer_data.Get_Correction_Data(dialog_data.description);

		text = d_thermometer_data.Get_Calibration_Due_Date(dialog_data.description).toString(QStringLiteral("MMM dd yyyy"));
		this->Add_Log_Text(QString("INF  Updated correction data for '%1'.  Calibration Due: %2").arg(dialog_data.description).arg(text));
		
		this->Update_All_Temperature_Corrections();
		
		d_thermometer_data.Save_Data(d_data_path + QStringLiteral("/calibration_data.xml"));
	}
}

void TTemperatureRecord::Remove_Thermometer(void)
{
	QStringList							description_list;
	QString								description;
	
	d_msg_box->setText("Remove Thermometer");
	d_msg_box->setInformativeText("Are you sure you want to delete the current thermometer?");
	d_msg_box->setDetailedText(QString());
	d_msg_box->setIcon(QMessageBox::Warning);
	d_msg_box->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
	d_msg_box->setDefaultButton(QMessageBox::No);
	
	if(QMessageBox::Yes == d_msg_box->exec())
	{
		description = d_thermometer_combo->currentText();
		d_thermometer_data.Remove_Item(description);
		
		description_list = d_thermometer_data.Description_List();
		d_thermometer_combo->clear();
		d_thermometer_combo->addItems(description_list);
		
		d_active_thermometer_usable = false;

		if(d_thermometer_combo->count())
		{
			d_thermometer_combo->setCurrentIndex(0);
		}
		else
		{
			this->Update_All_Temperature_Corrections();
		}
		
		this->Add_Log_Text(QString("INF  Removed correction data for '%1'.").arg(description));
		
		d_thermometer_data.Save_Data(d_data_path + QStringLiteral("/calibration_data.xml"));
	}
}

void TTemperatureRecord::Thermometer_Selection_Changed(
	int									index)
{
	QString								description;
	
	if(!(index < 0))
	{
		description = d_thermometer_combo->itemText(index);
		
		d_active_correction_data = d_thermometer_data.Get_Correction_Data(description);
		d_active_thermometer_usable = d_thermometer_data.Is_Calibration_Current(description);
		
		this->Update_All_Temperature_Corrections();
	}
}

void TTemperatureRecord::Temperature_Input_Changed(void)
{
	QLineEdit							*input_edit;
	QLineEdit							*output_edit;
	TVector2							pnt;
	double								temp_in;
	double								temp_out;
	bool								state;
	
	input_edit = qobject_cast<QLineEdit*>(this->sender());
	if(input_edit)
	{
		if(input_edit == d_meter_x_edit)
		{
			output_edit = d_corrected_x_edit;
		}
		else if(input_edit == d_meter_y_edit)
		{
			output_edit = d_corrected_y_edit;
		}
		else if(input_edit == d_meter_z_edit)
		{
			output_edit = d_corrected_z_edit;
		}
		else if(input_edit == d_meter_part_edit)
		{
			output_edit = d_corrected_part_edit;
		}
		else
		{
			return;
		}

		if(d_active_thermometer_usable && d_active_correction_data.Valid_Range())
		{
			temp_in = input_edit->text().toDouble(&state);

			if(state)
			{
				pnt = d_active_correction_data.Value(temp_in);
				temp_out = pnt.x - pnt.y;
				output_edit->setText(QString("%1").arg(temp_out,0,'f',2));
			}
			else
			{
				output_edit->setText(QStringLiteral("-----"));
			}
		}
		else
		{
			output_edit->setText(QStringLiteral("-----"));
		}
	}
}

void TTemperatureRecord::Write_Temperature_Entry(void)
{
	QFile								file;
	QString								file_name;
	QString								text;
	bool								add_header(false);
	
	file_name = QDir::fromNativeSeparators(d_record_path_edit->text());
	file_name += QString("/%1").arg(d_record_file_edit->text());
	
	file.setFileName(file_name);
	
	if(!file.exists())
	{
		add_header = true;
		this->Add_Log_Text(QStringLiteral("INF  Creating log file: '%1' for '%2'").arg(QDir::toNativeSeparators(file_name)).arg(d_description_edit->text()));
	}
	else
	{
		this->Add_Log_Text(QStringLiteral("INF  Appending log file: '%1' for '%2'").arg(QDir::toNativeSeparators(file_name)).arg(d_description_edit->text()));
	}
	
	if(!file.open(QIODevice::Append | QIODevice::Text))
	{
		this->Add_Log_Text(QStringLiteral("WAR  Cannot open temperature log file for writing"));
		return;
	}
	
	if(add_header)
	{
		text = QStringLiteral("Measurement Description,Thermometer,X Temp. C,Y Temp. C,Z Temp. C,Part Temp. C\n");
		file.write(text.toLatin1());
	}
	
	text = d_description_edit->text();
	file.write(text.toLatin1());
	file.write(",");

	text = d_thermometer_combo->currentText();
	file.write(text.toLatin1());
	file.write(",");
	
	if(d_temperature_x_check->isChecked())
	{
		text = d_corrected_x_edit->text();
		file.write(text.toLatin1());
	}

	file.write(",");

	if(d_temperature_y_check->isChecked())
	{
		text = d_corrected_y_edit->text();
		file.write(text.toLatin1());
	}

	file.write(",");

	if(d_temperature_z_check->isChecked())
	{
		text = d_corrected_z_edit->text();
		file.write(text.toLatin1());
	}

	file.write(",");

	if(d_temperature_part_check->isChecked())
	{
		text = d_corrected_part_edit->text();
		file.write(text.toLatin1());
	}
	
	file.write("\n");
	
	file.close();
}

void TTemperatureRecord::Evaluate_Corrections(void)
{
	QFile								file;
	QDate								date;
	QString								file_name;
	QString								description;
	QString								text;
	TVector2							pnt;
	double								pos,start,end;
	double								dval;
	static const double					INCREMENT(0.1);
	static const double					ZERO_EPSILON(0.001);
	
	file_name = QFileDialog::getSaveFileName(this, 
											QStringLiteral("Save File"),
											d_current_path + QStringLiteral("/evaluation.csv"),
											QStringLiteral("CSV data (*.csv *.txt)"));

	if(file_name.length())
	{
		file.setFileName(file_name);
		
		if(!file.open(QIODevice::WriteOnly | QIODevice::Text))
		{
			this->Add_Log_Text(QStringLiteral("ERR  Cannot open evaluation file for writing."));
			return;
		}
		
		description = d_thermometer_combo->currentText();

		text = QString("Thermometer:,%1\n").arg(description);
		file.write(text.toLatin1());
		
		date = d_thermometer_data.Get_Calibration_Due_Date(description);
		text = QString("Calibration Due:,%1\n").arg(date.toString(QStringLiteral("MMM dd yyyy")));
		file.write(text.toLatin1());
		
		pnt = d_active_correction_data.Point1();
		start = pnt.x;
		text = QString("Point 1:,%1,%2\n").arg(pnt.x,0,'f',4).arg(pnt.y,0,'f',4);
		file.write(text.toLatin1());
		
		pnt = d_active_correction_data.Point2();
		text = QString("Point 2:,%1,%2\n").arg(pnt.x,0,'f',4).arg(pnt.y,0,'f',4);
		file.write(text.toLatin1());
		
		pnt = d_active_correction_data.Point3();
		end = pnt.x;
		text = QString("Point 3:,%1,%2\n").arg(pnt.x,0,'f',4).arg(pnt.y,0,'f',4);
		file.write(text.toLatin1());
		
		text = QStringLiteral("Reference C,Corrected C,Deviation C\n");
		file.write(text.toLatin1());

		// round start and end to nearest 1 C beyond existing range
		
		start = floor(start);
		end = ceil(end);
		end += ZERO_EPSILON;
		
		pos = start;
		
		while(pos < end)
		{
			pnt = d_active_correction_data.Value(pos);
			dval = pnt.x - pnt.y;
			
			text = QString("%1,%2,%3\n").arg(pos,0,'f',4).arg(dval,0,'f',4).arg(dval - pos);
			file.write(text.toLatin1());
			
			pos += INCREMENT;
		};
		
		file.close();
		
		this->Add_Log_Text(QStringLiteral("INF  Evaluation file written."));
	}
}

void TTemperatureRecord::Evaluate_Uncertainty(void)
{
	QFile								file;
	QDate								date;
	QString								file_name;
	QString								description;
	QString								text;
	TVector2							pnt;
	TVector2							p1,p2,p3;
	TCorrectionData						correction_data;
	double								of1,of2,of3;
	double								pos,start,end;
	double								dval;
	double								uncertainty;
	int									mask;
	static const double					INCREMENT(1.0);
	static const double					ZERO_EPSILON(0.001);
	
	if(!d_input_dialog)
	{
		d_input_dialog = new TInputDialog(this);
		d_input_dialog->Reset(0.04);
	}
	
	if(QDialog::Accepted == d_input_dialog->exec())
	{
		uncertainty = d_input_dialog->Uc_Value();
		
		file_name = QFileDialog::getSaveFileName(this, 
											QStringLiteral("Save File"),
											d_current_path + QStringLiteral("/uncertainty.csv"),
											QStringLiteral("CSV data (*.csv *.txt)"));
		
		if(file_name.length())
		{
			file.setFileName(file_name);
			
			if(!file.open(QIODevice::WriteOnly | QIODevice::Text))
			{
				this->Add_Log_Text(QStringLiteral("ERR  Cannot open evaluation file for writing."));
				return;
			}
			
			description = d_thermometer_combo->currentText();

			text = QString("Thermometer:,%1\n").arg(description);
			file.write(text.toLatin1());
			
			date = d_thermometer_data.Get_Calibration_Due_Date(description);
			text = QString("Calibration Due:,%1,,,Temp. C ->\n").arg(date.toString(QStringLiteral("MMM dd yyyy")));
			file.write(text.toLatin1());
			
			p1 = d_active_correction_data.Point1();
			p2 = d_active_correction_data.Point2();
			p3 = d_active_correction_data.Point3();

			// round start and end to nearest 1 C beyond existing range
			
			start = floor(p1.x);
			end = ceil(p3.x);
			end += ZERO_EPSILON;
			
			// create reference temperature columns
			text = QStringLiteral("Set,Offset 1,Offset 2,Offset 3");
			file.write(text.toLatin1());

			pos = start;
			while(pos < end)
			{				
				text = QString(",%1").arg(pos,0,'f',4);
				file.write(text.toLatin1());
				
				pos += INCREMENT;
			};
			file.write("\n");
			
			// write every combination of uc errors
			// using a bit masks for combinations
			for(mask = 0;mask < 0b1000000;++mask)
			{
				// duplicates
				if(mask && !(mask & 0b111))
				{
					continue;
				}
				
				of1 = 0;
				of2 = 0;
				of3 = 0;
				
				if(mask & 0b001)
				{
					(mask & 0b1000) ? of1 -= uncertainty : of1 += uncertainty;
				}
				
				if(mask & 0b010)
				{
					(mask & 0b10000) ? of2 -= uncertainty : of2 += uncertainty;
				}
				
				if(mask & 0b100)
				{
					(mask & 0b100000) ? of3 -= uncertainty : of3 += uncertainty;
				}
				
				correction_data.Set_Points(p1 + TVector2(0,of1),p2 + TVector2(0,of2),p3 + TVector2(0,of3));
				
				
				// write temperature columns
				text = QString("%1,%2,%3,%4").arg(mask).arg(of1,0,'f',3).arg(of2,0,'f',3).arg(of3,0,'f',3);
				file.write(text.toLatin1());

				pos = start;
				while(pos < end)
				{				
					pnt = correction_data.Value(pos);
					dval = pnt.x - pnt.y;
				
					text = QString(",%1").arg(dval,0,'f',4);
					file.write(text.toLatin1());
					
					pos += INCREMENT;
				};
				file.write("\n");
			}

			
			file.close();
			
			this->Add_Log_Text(QStringLiteral("INF  Uncertainty file written."));
		}
	}
}

void TTemperatureRecord::closeEvent(
	QCloseEvent 						*event)
{
	d_settings->setValue("Current_Path", QDir::fromNativeSeparators(d_record_path_edit->text()));
	d_settings->setValue("Record_File", d_record_file_edit->text());
	
	d_settings->setValue("Active_Thermometer", d_thermometer_combo->currentIndex());
	
	d_settings->setValue("Description", d_description_edit->text());
	
	d_settings->setValue("Axis_X_Enabled", d_temperature_x_check->isChecked());
	d_settings->setValue("Axis_Y_Enabled", d_temperature_y_check->isChecked());
	d_settings->setValue("Axis_Z_Enabled", d_temperature_z_check->isChecked());
	d_settings->setValue("Axis_Part_Enabled", d_temperature_part_check->isChecked());
	
	d_settings->setValue("Meter_X_Value", d_meter_x_edit->text());
	d_settings->setValue("Meter_Y_Value", d_meter_y_edit->text());
	d_settings->setValue("Meter_Z_Value", d_meter_z_edit->text());
	d_settings->setValue("Meter_Part_Value", d_meter_part_edit->text());
	
	event->accept();
}

void TTemperatureRecord::Update_All_Temperature_Corrections(void)
{
	double								temp_in;
	double								temp_out;
	TVector2							pnt;
	bool								state;
	
	if(d_active_thermometer_usable && d_active_correction_data.Valid_Range())
	{
		// X
		temp_in = d_meter_x_edit->text().toDouble(&state);
		
		if(state)
		{
			pnt = d_active_correction_data.Value(temp_in);
			temp_out = pnt.x - pnt.y;
			d_corrected_x_edit->setText(QString("%1").arg(temp_out,0,'f',2));
		}
		else
		{
			d_corrected_x_edit->setText(QStringLiteral("-----"));
		}
		
		// y
		temp_in = d_meter_y_edit->text().toDouble(&state);
		
		if(state)
		{
			pnt = d_active_correction_data.Value(temp_in);
			temp_out = pnt.x - pnt.y;
			d_corrected_y_edit->setText(QString("%1").arg(temp_out,0,'f',2));
		}
		else
		{
			d_corrected_y_edit->setText(QStringLiteral("-----"));
		}
		
		// z
		temp_in = d_meter_z_edit->text().toDouble(&state);
		
		if(state)
		{
			pnt = d_active_correction_data.Value(temp_in);
			temp_out = pnt.x - pnt.y;
			d_corrected_z_edit->setText(QString("%1").arg(temp_out,0,'f',2));
		}
		else
		{
			d_corrected_z_edit->setText(QStringLiteral("-----"));
		}
		
		// part
		temp_in = d_meter_part_edit->text().toDouble(&state);
		
		if(state)
		{
			pnt = d_active_correction_data.Value(temp_in);
			temp_out = pnt.x - pnt.y;
			d_corrected_part_edit->setText(QString("%1").arg(temp_out,0,'f',2));
		}
		else
		{
			d_corrected_part_edit->setText(QStringLiteral("-----"));
		}
	}
	else
	{
		d_corrected_x_edit->setText(QStringLiteral("-----"));
		d_corrected_y_edit->setText(QStringLiteral("-----"));
		d_corrected_z_edit->setText(QStringLiteral("-----"));
		d_corrected_part_edit->setText(QStringLiteral("-----"));
	}

}

void TTemperatureRecord::Clear_Log(void)
{
	d_log_edit->clear();
}

void TTemperatureRecord::Add_Log_Text(
	const QString 						&text)
{
	d_log_edit->appendPlainText(text);
	d_log_edit->moveCursor(QTextCursor::End);
}


