/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QFile>
#include <QByteArray>
#include <assert.h>

#include "../../core/xmlfile.h"

#include "thermometerdata.h"

TThermometerData::TThermometerData(void)
{
	d_checksum_error_count = 0;
}

TThermometerData::~TThermometerData(void)
{
}

QStringList TThermometerData::Description_List(void) const
{
	std::vector<TThermometerDataItem>::const_iterator iter;
	QStringList							list;
	
	for(iter = d_thermometer_data_items.begin();iter != d_thermometer_data_items.end();++iter)
	{
		list.push_back((*iter).description);
	}
	
	return list;
}

bool TThermometerData::Get_Temperature_Data(
	const QString						&description,
	TTemperatureData					* const temperature_data) const
{
	std::vector<TThermometerData::TThermometerDataItem>::const_iterator iter;
	TVector2							pnt;
	
	assert(temperature_data);

	// defaults
	temperature_data->reference_temperature_1 = 15.0;
	temperature_data->reference_temperature_2 = 20.0;
	temperature_data->reference_temperature_3 = 25.0;
	temperature_data->corrected_temperature_1 = 15.0;
	temperature_data->corrected_temperature_2 = 20.0;
	temperature_data->corrected_temperature_3 = 25.0;
	
	if(description.length())
	{
		for(iter = d_thermometer_data_items.begin();iter != d_thermometer_data_items.end();++iter)
		{
			if(description.compare((*iter).description,Qt::CaseInsensitive) == 0)
			{
				pnt = (*iter).correction_data.Point1();
				temperature_data->reference_temperature_1 = pnt.x;
				temperature_data->corrected_temperature_1 = pnt.x + pnt.y;

				pnt = (*iter).correction_data.Point2();
				temperature_data->reference_temperature_2 = pnt.x;
				temperature_data->corrected_temperature_2 = pnt.x + pnt.y;

				pnt = (*iter).correction_data.Point3();
				temperature_data->reference_temperature_3 = pnt.x;
				temperature_data->corrected_temperature_3 = pnt.x + pnt.y;
				
				return true;
			}
		}
	}
	
	return false;
}

TCorrectionData TThermometerData::Get_Correction_Data(
	const QString						&description) const
{
	std::vector<TThermometerData::TThermometerDataItem>::const_iterator iter;

	if(description.length())
	{
		for(iter = d_thermometer_data_items.begin();iter != d_thermometer_data_items.end();++iter)
		{
			if(description.compare((*iter).description,Qt::CaseInsensitive) == 0)
			{
				return (*iter).correction_data;
			}
		}
	}
	
	return TCorrectionData();
}

QDate TThermometerData::Get_Calibration_Due_Date(
	const QString						&description) const
{
	std::vector<TThermometerData::TThermometerDataItem>::const_iterator iter;
	
	if(description.length())
	{
		for(iter = d_thermometer_data_items.begin();iter != d_thermometer_data_items.end();++iter)
		{
			if(description.compare((*iter).description,Qt::CaseInsensitive) == 0)
			{
				return (*iter).calibration_due_date;
			}
		}
	}
	
	return QDate::currentDate();
}

bool TThermometerData::Is_Calibration_Current(
	const QString						&description) const
{
	std::vector<TThermometerData::TThermometerDataItem>::const_iterator iter;
	
	if(description.length())
	{
		for(iter = d_thermometer_data_items.begin();iter != d_thermometer_data_items.end();++iter)
		{
			if(description.compare((*iter).description,Qt::CaseInsensitive) == 0)
			{
				return ((*iter).calibration_due_date > QDate::currentDate());
			}
		}
	}
	
	return false;
}

bool TThermometerData::Save_Data(
	const QString 						&file_name) const
{
	std::vector<TThermometerData::TThermometerDataItem>::const_iterator iter;
	TXmlFile							xml_file;
	QDomElement							e0;
	QDomElement							e1;
	QDomElement							e2;
	TVector2							pnt;
	QString								text;
	
	try
	{
		xml_file.Set_File_Name(file_name);
		xml_file.Reset("Thermometer_Calibration_Data",1);

		// temp data entries
		e0 = xml_file.Create_Node(xml_file.Root_Node(),QStringLiteral("Calibration_Entries"));

		for(iter = d_thermometer_data_items.begin();iter != d_thermometer_data_items.end();++iter)
		{
			e1 = xml_file.Create_Node(&e0,QStringLiteral("Entry"));

			e2 = xml_file.Create_Node(&e1,QStringLiteral("Description"));
			xml_file.Write_Text_Node(&e2,(*iter).description);
			
			e2 = xml_file.Create_Node(&e1,QStringLiteral("Calibration_Due_Date"));
			xml_file.Write_Text_Node(&e2,(*iter).calibration_due_date.toString(QStringLiteral("ddMMyyyy")));

			e2 = xml_file.Create_Node(&e1,QStringLiteral("Point1"));
			pnt = (*iter).correction_data.Point1();
			xml_file.Write_Text_Node(&e2,QString("%1,%2").arg(pnt.x,0,'f',5).arg(pnt.y,0,'f',5));

			e2 = xml_file.Create_Node(&e1,QStringLiteral("Point2"));
			pnt = (*iter).correction_data.Point2();
			xml_file.Write_Text_Node(&e2,QString("%1,%2").arg(pnt.x,0,'f',5).arg(pnt.y,0,'f',5));

			e2 = xml_file.Create_Node(&e1,QStringLiteral("Point3"));
			pnt = (*iter).correction_data.Point3();
			xml_file.Write_Text_Node(&e2,QString("%1,%2").arg(pnt.x,0,'f',5).arg(pnt.y,0,'f',5));
			
			e2 = xml_file.Create_Node(&e1,QStringLiteral("Checksum"));
			text = this->Checksum(*iter);
			xml_file.Write_Text_Node(&e2,text);
		}
	
		xml_file.Write_File();
	}
	catch(TXmlFileException XmlFileException)
	{
		d_last_error = QString("error writing file '%1': %2").arg(file_name).arg(XmlFileException.ErrorText());
		return false;
	}
	
	return true;
}
	
bool TThermometerData::Load_Data(
	const QString 						&file_name)
{
	TXmlFile							xmlfile;
	int									version;
		
	try
	{
		xmlfile.Set_File_Name(file_name);
		xmlfile.Open("Thermometer_Calibration_Data",&version);
		
		switch(version)
		{
			case 1:
				this->Load_Data_V1(&xmlfile);
				break;
		}
	}
	catch(TXmlFileException xml_exception)
	{
		d_last_error = QString("error reading file '%1': %2").arg(file_name).arg(xml_exception.ErrorText());
		return false;
	}
		
	return true;
}

void TThermometerData::Set_Temperature_Data(
	const QString						&description,
	const TTemperatureData				&temperature_data)
{
	std::vector<TThermometerData::TThermometerDataItem>::iterator iter;
	TThermometerData::TThermometerDataItem thermometer_data_item;
	TVector2							p1,p2,p3;

	if(description.length())
	{
		p1.Set(temperature_data.reference_temperature_1,temperature_data.corrected_temperature_1 - temperature_data.reference_temperature_1);
		p2.Set(temperature_data.reference_temperature_2,temperature_data.corrected_temperature_2 - temperature_data.reference_temperature_2);
		p3.Set(temperature_data.reference_temperature_3,temperature_data.corrected_temperature_3 - temperature_data.reference_temperature_3);

		for(iter = d_thermometer_data_items.begin();iter != d_thermometer_data_items.end();++iter)
		{
			if(description.compare((*iter).description,Qt::CaseInsensitive) == 0)
			{
				(*iter).correction_data.Set_Points(p1,p2,p3);
				return;
			}
		}
		
		thermometer_data_item.description = description;
		thermometer_data_item.correction_data.Set_Points(p1,p2,p3);
		
		d_thermometer_data_items.push_back(thermometer_data_item);
	}
}

void TThermometerData::Set_Calibration_Due_Date(
	const QString						&description,
	const QDate							&date)
{
	std::vector<TThermometerData::TThermometerDataItem>::iterator iter;
	
	if(description.length())
	{
		for(iter = d_thermometer_data_items.begin();iter != d_thermometer_data_items.end();++iter)
		{
			if(description.compare((*iter).description,Qt::CaseInsensitive) == 0)
			{
				(*iter).calibration_due_date = date;
				return;
			}
		}
	}
}

void TThermometerData::Remove_Item(
	const QString 						&description)
{
	std::vector<TThermometerData::TThermometerDataItem>::iterator iter;

	if(description.length())
	{
		for(iter = d_thermometer_data_items.begin();iter != d_thermometer_data_items.end();++iter)
		{
			if(description.compare((*iter).description,Qt::CaseInsensitive) == 0)
			{
				d_thermometer_data_items.erase(iter);
				return;
			}
		}
	}
}

QString TThermometerData::Checksum(
	const TThermometerDataItem			&thermometer_data_item) const
{
	QString								text;
	QByteArray							data;
	TVector2							p1,p2,p3;
	unsigned short						sum;
	int									cntr;
	
	text = thermometer_data_item.description;
	
	p1 = thermometer_data_item.correction_data.Point1();
	p2 = thermometer_data_item.correction_data.Point2();
	p3 = thermometer_data_item.correction_data.Point3();
	
	text += thermometer_data_item.calibration_due_date.toString(QStringLiteral("ddMMyyyy"));

	text += QString("%1%2").arg(p1.x,0,'f',2).arg(p1.y,0,'f',4);
	text += QString("%1%2").arg(p2.x,0,'f',2).arg(p2.y,0,'f',4);
	text += QString("%1%2").arg(p3.x,0,'f',2).arg(p3.y,0,'f',4);
	
	
	data = text.toUtf8();
	
	sum = 0;
	for(cntr = 0;cntr < data.size();++cntr)
	{
		sum += static_cast<unsigned short>(data.at(cntr));
	}
	
	text = QString("%1").arg(sum,4,16,QChar('0'));
	
	return text;
}

void TThermometerData::Load_Data_V1(
	TXmlFile							* const xml_file)
{
	QDomElement							e0;
	QDomElement							e1;
	QDomElement							e2;
	QString								text;
	QStringList							list;
	TVector2							p1,p2,p3;
	TThermometerData::TThermometerDataItem data_item;

	assert(xml_file);
	
	d_thermometer_data_items.clear();
	d_checksum_error_count = 0;

	e0 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Calibration_Entries"));
	e1 = xml_file->Get_Node(&e0,QStringLiteral("Entry"),true);
	
	while(!e1.isNull())
	{
		e2 = xml_file->Get_Node(&e1,QStringLiteral("Description"));
		data_item.description = xml_file->Read_Text_Node(&e2);

		e2 = xml_file->Get_Node(&e1,QStringLiteral("Calibration_Due_Date"));
		data_item.calibration_due_date = QDate::fromString(xml_file->Read_Text_Node(&e2),QStringLiteral("ddMMyyyy"));

		e2 = xml_file->Get_Node(&e1,QStringLiteral("Point1"));
		text = xml_file->Read_Text_Node(&e2);
		list = text.split(',');
		
		if(list.size() != 2)
		{
			throw TXmlFileException(QStringLiteral("Point1 temperature data invalid"));
		}
		
		p1.x = list[0].toDouble();
		p1.y = list[1].toDouble();

		e2 = xml_file->Get_Node(&e1,QStringLiteral("Point2"));
		text = xml_file->Read_Text_Node(&e2);
		list = text.split(',');
		
		if(list.size() != 2)
		{
			throw TXmlFileException(QStringLiteral("Point2 temperature data invalid"));
		}
		
		p2.x = list[0].toDouble();
		p2.y = list[1].toDouble();
		
		e2 = xml_file->Get_Node(&e1,QStringLiteral("Point3"));
		text = xml_file->Read_Text_Node(&e2);
		list = text.split(',');
		
		if(list.size() != 2)
		{
			throw TXmlFileException(QStringLiteral("Point3 temperature data invalid"));
		}
		
		p3.x = list[0].toDouble();
		p3.y = list[1].toDouble();
		
		data_item.correction_data.Set_Points(p1,p2,p3);

		e2 = xml_file->Get_Node(&e1,QStringLiteral("Checksum"));
		text = xml_file->Read_Text_Node(&e2);
		
		if(text.compare(this->Checksum(data_item)) == 0)
		{
			d_thermometer_data_items.push_back(data_item);
		}
		else
		{
			++d_checksum_error_count;
		}

		e1 = xml_file->Get_Sibling_Node(&e1,QStringLiteral("Entry"),true);
	};
}
